"""
profile basic epub generation and return stats
"""
import os
import pstats
import logging
import unittest
import cProfile

from .. import Epub, create_chapter_from_url

#** Functions **#

def _create_epub():
    try:
        epub    = Epub('My First Epub')
        chapter = create_chapter_from_url('https://en.wikipedia.org/wiki/Grand_Teton_National_Park')
        epub.add_chapter(chapter)
        epub.create('./profile-test.epub')
    finally:
        try:
            os.remove('./profile-test.epub')
        except Exception:
            pass

#** Tests **#

class Profiling(unittest.TestCase):
    """Profile Based UnitTests"""

    def test_epub_profile(self):
        """complete profile on generating the given epub"""
        # enable profile
        pr = cProfile.Profile()
        pr.enable()
        # generate stats
        ps = pstats.Stats(pr, stream=_create_epub()).sort_stats('cumulative')
        ps.print_stats()
        # disable profile
        pr.disable()
        self.assertEqual(True, True)

#** Main **#

if __name__ == '__main__':
    logging.getLogger('pypub').setLevel(logging.DEBUG)
    unittest.main()



import os
import unittest

from ..chapter import *

#** Variables **#
__all__ = ['ChapterTests']

#: static directory for unit-test related resources
STATIC = os.path.realpath(os.path.join(os.path.dirname(__file__), 'static'))

#** Tests **#

class ChapterTests(unittest.TestCase):
 
    def test_create_chapter_from_text(self):
        """complete test for create_chapter_from_text"""
        c = create_chapter_from_text('Hello\nWorld!\n\n')
        self.assertIn(b'<p>Hello</p>', c.content)
        self.assertIn(b'<p>World!</p>', c.content)

    def test_create_chapter_from_file(self):
        """complete test for create_chapter_from_file"""
        example = os.path.join(STATIC, 'example.html')
        c       = create_chapter_from_file(example)
        self.assertEqual(c.title, 'Business & Co')
        self.assertEqual(c.url, f'file://{os.path.abspath(example)}')

    def test_create_chapter_from_url(self):
        """complete test for create_chapter_from_url"""
        c = create_chapter_from_url('https://www.example.com')
        self.assertEqual(c.title, 'Example Domain')
        self.assertEqual(c.url, 'https://www.example.com')
        self.assertIn(b'illustrative examples', c.content)

    def test_create_chapter_example(self):
        """complete test for readme example"""
        c = create_chapter_from_url('https://en.wikipedia.org/wiki/EPUB')
        self.assertEqual(c.title, 'EPUB - Wikipedia')
        self.assertEqual(c.url, 'https://en.wikipedia.org/wiki/EPUB')
        self.assertIn(b'EPUB 3.2 was announced in 2018', c.content)

#** Main **#

if __name__ == '__main__':
    unittest.main()



"""
Chapter Rendering Factory Implementation
"""
import os
import uuid
import urllib.error
import urllib.parse
import urllib.request
from logging import Logger
from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Protocol, Optional, cast

import filetype
import pyxml.html
from pyxml.html import HtmlElement
from jinja2 import Template

from .chapter import Chapter, urlrequest, htmltostring

#** Variables **#
__all__ = [
    'SUPPORTED_TAGS',

    'urlrequest',
    'render_images',
    'xmlprettify',

    'RenderCtx',
    'ChapterFactory',
    'SimpleChapterFactory'
]

#NOTE: monkey patch pyxml to not use &nbsp escape character in attributes
from pyxml.escape import ESCAPE_ATTRIB
ESCAPE_ATTRIB.pop(' ', None)

#: unicode characters to replace if found in content
REPLACE = dict([(ord(x), ord(y)) for x,y in zip(u"‘’´“”–-", u"'''\"\"--")])

#: supported tags and attributes allowed in an epub
SUPPORTED_TAGS = {
    'a':          ('href', 'id'),
    'b':          ('id', ),
    'big':        (),
    'blockquote': ('id', ),
    'body':       (),
    'br':         ('id', ),
    'center':     (),
    'cite':       (),
    'dd':         ('id', 'title'),
    'del':        (),
    'dfn':        (),
    'div':        ('align', 'id', 'bgcolor'),
    'em':         ('id', 'title'),
    'font':       ('color', 'face', 'id', 'size'),
    'head':       (),
    'h1':         (),
    'h2':         (),
    'h3':         (),
    'h4':         (),
    'h5':         (),
    'h6':         (),
    'html':       (),
    'i':          ('class', 'id'),
    'img':        ('align', 'border', 'height', 'id', 'src', 'width'),
    'li':         ('class', 'id', 'title'),
    'ol':         ('id', ),
    'p':          ('align', 'id', 'title'),
    's':          ('id', 'style', 'title'),
    'small':      ('id', ),
    'span':       ('bgcolor', 'title'),
    'strike':     ('class', 'id'),
    'strong':     ('class', 'id'),
    'sub':        ('id', ),
    'sup':        ('class', 'id'),
    'table':      ('class', 'id'),
    'th':         ('class', 'id'),
    'tr':         ('class', 'id'),
    'td':         ('class', 'id'),
    'u':          ('id', ),
    'ul':         ('class', 'id'),
    'var':        (),
}

#** Functions **#

def mime_type(url: str, data: bytes) -> Optional[str]:
    """
    determine mimetype of the given file w/ first chunk of data

    :param url:  download url
    :param data: first n-bytes of file
    :return:     determined mime-type (if found)
    """
    mime = filetype.image_match(data)
    if mime is not None:
        return mime.EXTENSION
    # support svg images
    if url.endswith('svg') and b'<svg' in data:
        return 'svg'

def render_images(ctx: 'RenderCtx', chunk_size: int = 8192):
    """
    replace global image references w/ local downloaded ones
    """
    downloads = {}
    for image in ctx.etree.xpath('.//img[@src]'):
        # cleanup link and resolve relative paths
        url = image.attrib['src'].rsplit('?', 1)[0]
        fmt = (ctx.chapter.title, url)
        if '://' not in url:
            if not ctx.chapter.url:
                ctx.logger.warning(
                    'chapter[%s] cannot render image %r w/o chapter-url' % fmt)
                continue
            url = urllib.parse.urljoin(ctx.chapter.url, urllib.parse.quote(url))
        else:
            # make sure the url is properly encoded
            url = urllib.parse.quote(url, safe=":/")
        # skip if url has already been downloaded
        if url in downloads:
            image.attrib['src'] = downloads[url]
            continue
        # download url into local image folder for epub
        ctx.logger.debug(f'chapter[%s] downloading image: %r' % fmt)
        try:
            res = urlrequest(url, timeout=ctx.timeout)
            # ensure status of response is valid
            status = getattr(res, 'status', None)
            if status and status != 200:
                raise urllib.error.URLError(f'status: {status}')
            # read first chunk to determine content-type
            chunk = res.read(chunk_size)
            mime  = mime_type(url, chunk)
            if not mime:
                ctx.logger.warning('chapter[%s] cannot identify %r mime' % fmt)
                continue
            fname = f'image-{uuid.uuid4()}.{mime}'
            fpath = os.path.join(ctx.imagedir, fname)
            # read rest of the content into associated file
            with open(fpath, 'wb') as f:
                while chunk:
                    f.write(chunk)
                    chunk = res.read(chunk_size)
            # save epub-path in downloads cache and update image attribs
            epub_path = os.path.join('images/', fname)
            downloads[url] = epub_path
            image.attrib['src'] = epub_path
        except urllib.error.URLError:
            ctx.logger.error('chapter[%s] failed to download %r' % fmt)

def externalize_links(url: str, elem: HtmlElement):
    """
    iterate every anchor tag and externalize the links when able

    :param url:  url for chapter being rendered
    :param elem: root element for chapter object
    """
    for e in elem.finditer('//a'):
        href = e.attrib.get('href', None)
        if not href or '://' in href[:10] or href.startswith('#'):
            continue
        e.attrib['href'] = urllib.parse.urljoin(url, href)

def xmlprettify(elem: HtmlElement, chars: str='  ', level: int=1):
    """
    prettify the given element-tree w/ new indentations

    :param elem:  element being prettified
    :param chars: characters used for single indent
    :param level: internal variable used to track indent level
    """
    start    = '\n' + level * chars
    end      = '\n' + (level - 1) * chars
    children = elem.getchildren()
    if children:
        # make modifications
        if elem.text:
            elem.text = elem.text.rstrip() + start
        if elem.tail:
            elem.tail = elem.tail.rstrip() + end
        # iterate children
        for child in children:
            xmlprettify(child, chars, level + 1)
        # ensure last child has tail indentation of parent
        if children:
            child = children[-1]
            child.tail = (child.tail or '').rstrip() + end
    else:
        elem.text = '' if not elem.text else elem.text
        if elem.tail:
            elem.tail = (elem.tail or '').rstrip() + end

#** Classes **#

@dataclass
class RenderCtx:
    """variables and data associated w/ rendering a chapter"""
    logger:        Logger
    chapter:       Chapter
    etree:         HtmlElement
    imagedir:      str
    template:      Template
    render_kwargs: dict = field(default_factory=dict)
    extern_links:  bool = False
    timeout:       int  = 10

class ChapterFactory(Protocol):

    @abstractmethod
    def cleanup_html(self, content: bytes) -> HtmlElement:
        """
        cleanup raw html content and render as an lxml element tree
        """
        raise NotImplementedError

    @abstractmethod
    def hydrate(self, ctx: RenderCtx):
        """
        modify contents of chapter and etree before final rendering
        """
        raise NotImplementedError

    @abstractmethod
    def finalize(self, ctx: RenderCtx) -> bytes:
        """
        complete final rendering of chapter object and render as html bytes
        """
        raise NotImplementedError

    @abstractmethod
    def prettify(self, root: HtmlElement):
        """
        factory implementation to prettify etree at the end of finalization
        """
        raise NotImplementedError

    def render(self, log: Logger, chapter: Chapter, *args, **kwargs) -> bytes:
        """
        render chapter to bytes w/ the given settings
        """
        etree = self.cleanup_html(chapter.content)
        ctx   = RenderCtx(log, chapter, etree, *args, **kwargs)
        self.hydrate(ctx)
        return self.finalize(ctx)

class SimpleChapterFactory(ChapterFactory):
    """Simple Chapter Factory Implementation"""

    def cleanup_html(self, content: bytes) -> HtmlElement:
        """
        cleanup html content to only include supported tags
        """
        content = content.decode('utf-8', 'replace').translate(REPLACE).encode()
        # check if we can minimalize the scope
        etree   = pyxml.html.fromstring(content)
        body    = etree.xpath('.//body')
        etree   = body[0] if body else etree
        article = etree.xpath('.//article')
        etree   = article[0] if article else etree
        # iterate elements in tree and delete/modify them
        for elem in [elem for elem in etree.iter()][1:]:
            # if element tag is supported
            if elem.tag in SUPPORTED_TAGS:
                # remove attributes not approved for specific tag or no value
                for attr, value in list(elem.attrib.items()):
                    if attr not in SUPPORTED_TAGS[elem.tag] or not value:
                        elem.attrib.pop(attr)
            # if element is not supported, append children to parent
            else:
                # retrieve parent
                parent = cast(HtmlElement, elem.getparent())
                for child in elem.getchildren():
                    parent.append(child)
                parent.remove(elem)
                #NOTE: this is a bug with lxml, some children have
                # text in the parent included in the tail rather
                # than text attribute, so we also append tail to text
                if elem.tail and elem.tail.strip():
                    parent.text = (parent.text or '') + elem.tail.strip()
        # fix and remove invalid images
        for img in etree.xpath('.//img'):
            # ensure all images with no src are removed
            if 'src' not in img.attrib:
                cast(HtmlElement, img.getparent()).remove(img)
            # ensure they also have an alt
            elif 'alt' not in img.attrib:
                img.attrib['alt'] = img.attrib['src']
        # return new element-tree
        return etree

    def hydrate(self, ctx: RenderCtx):
        """
        modify chapter element-tree to render images
        """
        render_images(ctx)
        if ctx.extern_links and ctx.chapter.url:
            externalize_links(ctx.chapter.url, ctx.etree)

    def prettify(self, root: HtmlElement):
        """
        simple xml prettify function
        """
        xmlprettify(root)

    def finalize(self, ctx: RenderCtx) -> bytes:
        """
        render chapter content w/ specified template
        """
        content = ctx.template.render(**ctx.render_kwargs)
        # attach elements from chapter etree to content etree
        etree = pyxml.html.fromstring(content.encode())
        body  = etree.xpath('.//body')[0]
        for elem in ctx.etree.getchildren():
            body.append(elem)
        # return html as string to be written
        self.prettify(etree)
        return htmltostring(etree)



"""
Epub Book Generator
"""
import os
import uuid
import shutil
import logging
import zipfile
import tempfile
import mimetypes
from logging import Logger
from datetime import datetime
from dataclasses import dataclass, field
from typing import NamedTuple, Optional, Tuple, List

from PIL import Image, ImageDraw, ImageFont
from jinja2 import Environment, FileSystemLoader, Template

from .chapter import Chapter
from .factory import ChapterFactory, SimpleChapterFactory

#** Variables **#
__all__ = ['AssignedChapter', 'Assignment', 'EpubSpec', 'EpubBuilder']

#: assigned chapter type definition
AssignedChapter = Tuple['Assignment', Chapter]

#: base library directory
BASE = os.path.dirname(__file__)

#: static directory path
STATIC = os.path.join(BASE, 'static/')

#: templates directory to render content from
TEMPLATES = os.path.join(BASE, 'templates/')

#: base image used to generate cover
COVER_IMG = os.path.join(STATIC, 'img/cover.png')

#: font used to generater cover text
COVER_FONT = os.path.join(STATIC, 'fonts/free_mono.ttf')

#: jinja2 environment to load templates
jinja_env = Environment(loader=FileSystemLoader(TEMPLATES))
jinja_env.filters['basename'] = os.path.basename

#: default logging instance for library
default_logger = logging.getLogger('pypub')

#** Functions **#

def epub_dirs(basedir: Optional[str] = None) -> 'EpubDirs':
    """generate directories for epub data"""
    basedir = basedir or tempfile.mkdtemp()
    oebps   = os.path.join(basedir, 'OEBPS')
    metainf = os.path.join(basedir, 'META-INF')
    images  = os.path.join(oebps, 'images')
    styles  = os.path.join(oebps, 'styles')
    os.makedirs(oebps)
    os.makedirs(metainf)
    os.makedirs(images)
    os.makedirs(styles)
    return EpubDirs(basedir, oebps, metainf, images, styles)

def copy_file(src: str, into: str):
    """copy filepath into the `into` directory"""
    fname = os.path.basename(src)
    dst   = os.path.join(into, fname)
    shutil.copyfile(src, dst)

def copy_static(fpath: str, into: str):
    """copy static filepath into the `into` directory"""
    src = os.path.join(STATIC, fpath)
    copy_file(src, into)

def generate_font(text: str, target_ratio: float, maxsize: int, maxwidth: int):
    """generate font that fits the size of the image w/o overflow"""
    size = 5
    font = ImageFont.truetype(COVER_FONT, 5)
    # iterate until the text size is just larger than the criteria
    while font.getlength(text) <= target_ratio*maxwidth and size < maxsize:
        size += 1
        font = ImageFont.truetype(COVER_FONT, size)
    return font

def get_textsize(draw, text, font) -> Tuple[int, int]:
    """retrieve text-size of the current draw image"""
    if hasattr(draw, 'textbbox'):
        return draw.textbbox((0, 0), text, font=font)[2:]
    return draw.textsize(text, font=font)

def generate_cover(title: str, author: str, images: str) -> str:
    """generate cover image using PIL text overlay"""
    title  = title.title()
    author = author.title()
    fill   = (255, 255, 255, 255)
    with Image.open(COVER_IMG).convert("RGBA") as image:
        width, height = image.size
        draw = ImageDraw.Draw(image)
        # draw title on the top of the cover
        font = generate_font(title, 0.9, 60, width)
        w, _ = get_textsize(draw, title, font=font)
        draw.text(((width-w)/2, 10), title, font=font, fill=fill)
        # draw author on the bottom of the cover
        font = generate_font(title, 0.5, 40, width)
        w, h = get_textsize(draw, author, font=font)
        draw.text(((width-w)/2, height-int(h*1.5)), author, font=font, fill=fill)
        # write cover directly into epub directory
        fpath = os.path.join(images, 'cover.png')
        image.save(fpath)
        return 'cover.png'

def image_mime(fname: str) -> str:
    """get image mime of the given filename"""
    (ext, _) = mimetypes.guess_type(fname)
    return ext or 'image/unknown'

def render_template(name: str, into: str, encoding: str, kwargs: dict):
    """render template in templates directory"""
    fname    = name.rsplit('.j2', 1)[0]
    dest     = os.path.join(into, os.path.basename(fname))
    template = jinja_env.get_template(name)
    content  = template.render(**kwargs)
    with open(dest, 'w', encoding=encoding) as f:
        f.write(content)

#** Classes **#

@dataclass
class Assignment:
    id:         str
    link:       str
    play_order: int

@dataclass
class EpubDirs:
    basedir: str
    oebps:   str
    metainf: str
    images:  str
    styles:  str

class MimeFile(NamedTuple):
    name: str
    mime: str

@dataclass
class EpubSpec:
    """
    Epub Builder Specification
    """
    title:        str
    creator:      str            = 'pypub'
    language:     str            = 'en'
    rights:       str            = ''
    publisher:    str            = 'pypub'
    encoding:     str            = 'utf-8'
    date:         datetime       = field(default_factory=datetime.now)
    cover:        Optional[str]  = None
    epub_dir:     Optional[str]  = None
    extern_links: bool           = False
    factory:      ChapterFactory = field(repr=False, default_factory=SimpleChapterFactory)
    logger:       Logger         = field(repr=False, default=default_logger)
    css_paths:    List[str]      = field(repr=False, default_factory=list)

class EpubBuilder:
    """
    Epub Builder Class for Constructing Epub Books
    """

    def __init__(self, epub: EpubSpec):
        self.uid      = str(uuid.uuid4())
        self.epub     = epub
        self.logger   = epub.logger
        self.factory  = epub.factory
        self.encoding = epub.encoding
        self.dirs:     Optional[EpubDirs]    = None
        self.cover:    Optional[str]         = None
        self.template: Optional[Template]    = None
        self.chapters: List[AssignedChapter] = []

    def __enter__(self):
        """begin epub building in context-manager"""
        self.begin()
        return self

    def __exit__(self, *_):
        """cleanup on exit of context-manager"""
        self.cleanup()

    def begin(self) -> EpubDirs:
        """begin building operations w/ basic file structure"""
        # read jinja2 chapter template
        if not self.template:
            self.template = jinja_env.get_template('page.xhtml.j2')
        if self.dirs:
            return self.dirs
        args = (self.epub.title, self.epub.creator)
        self.logger.info('generating: %r (by: %s)' % args)
        # generate base directories and copy static files
        self.dirs = epub_dirs(self.epub.epub_dir)
        copy_static('mimetype', self.dirs.basedir)
        copy_static('container.xml', self.dirs.metainf)
        copy_static('css/coverpage.css', self.dirs.styles)
        copy_static('css/styles.css', self.dirs.styles)
        for path in self.epub.css_paths:
            copy_file(path, self.dirs.styles)
        # generate cover-image
        if self.epub.cover is not None:
            self.cover = os.path.basename(self.epub.cover)
            copy_file(self.epub.cover, self.dirs.images)
        else:
            self.logger.info('generating cover-image (%r by %r)' % args)
            self.cover = generate_cover(*args, self.dirs.images)
        # render cover-image
        fpath    = os.path.join(self.dirs.oebps, 'coverpage.xhtml')
        template = jinja_env.get_template('coverpage.xhtml.j2')
        cover    = f'images/{self.cover}'
        with open(fpath, 'w', encoding=self.encoding) as f:
            cover = template.render(cover=cover, epub=self.epub)
            f.write(cover)
        return self.dirs

    def render_chapter(self, assign: Assignment, chapter: Chapter):
        """render an assigned chapter into the ebook"""
        if not self.dirs or not self.template:
            raise RuntimeError('cannot render_chapter before `begin`')
        # log chapter generation
        self.chapters.append((assign, chapter))
        self.logger.info('rendering chapter #%d: %r' % (
            assign.play_order, chapter.title))
        # render chapter w/ appropriate kwargs
        args    = (self.logger, chapter, self.dirs.images, self.template)
        kwargs  = {'epub': self.epub, 'chapter': chapter}
        fpath   = os.path.join(self.dirs.oebps, assign.link)
        content = self.factory.render(*args, kwargs,
            extern_links=self.epub.extern_links)
        with open(fpath, 'wb') as f:
            f.write(content)

    def index(self):
        """build index files for epub before finalizing"""
        if not self.dirs or not self.cover:
            raise RuntimeError('cannot index epub before `begin`')
        kwargs = {
            'uid':      self.uid,
            'epub':     self.epub,
            'cover':    MimeFile(self.cover, image_mime(self.cover)),
            'styles':   os.listdir(self.dirs.styles),
            'chapters': self.chapters,
            'images':   [
                MimeFile(fname, image_mime(fname))
                for fname in os.listdir(self.dirs.images)
                if fname != self.cover
            ]
        }
        # render and write the rest of the templates
        self.logger.info('epub=%r, writing final templates' % self.epub.title)
        render_template('book.ncx.j2', self.dirs.oebps, self.encoding, kwargs)
        render_template('book.opf.j2', self.dirs.oebps, self.encoding, kwargs)
        render_template('toc.xhtml.j2', self.dirs.oebps, self.encoding, kwargs)

    def compress(self, fpath: Optional[str] = None) -> str:
        """compress build and zip content into epub"""
        if not self.dirs:
            raise RuntimeError('cannot finalize before `begin`')
        # reformat and build fpath w/ defaults
        fname = os.path.basename(fpath) if fpath else self.epub.title
        fname = fname.rsplit('.epub', 1)[0] + '.epub'
        fpath = os.path.dirname(fpath) if fpath else '.'
        fpath = os.path.join(fpath, fname)
        fzip  = fpath.rsplit('.epub', 1)[0] + '.zip'
        # zip contents of directory
        self.logger.info('epub=%r, zipping content' % self.epub.title)
        zipf = zipfile.ZipFile(fzip, 'w', zipfile.ZIP_DEFLATED)
        for root, _, files in os.walk(self.dirs.basedir):
            relpath = root.split(self.dirs.basedir, 1)[1]
            for file in files:
                real   = os.path.join(root, file)
                local  = os.path.join(relpath, file)
                method = zipfile.ZIP_STORED if file == 'mimetype' else None
                zipf.write(real, local, method)
        # rename zip to epub
        zipf.close()
        os.rename(fzip, fpath)
        return fpath

    def finalize(self, fpath: Optional[str] = None) -> str:
        """finalize build and index and compress epub directory"""
        self.index()
        return self.compress(fpath)

    def cleanup(self):
        """cleanup leftover resources after finalization"""
        if self.dirs:
            shutil.rmtree(self.dirs.basedir, ignore_errors=True)
            self.dirs = None



import os
import unittest
from typing import Set

from . import LOCAL_IMAGE, STATIC
from ..epub import Epub
from ..chapter import create_chapter_from_file

#** Variables **#
__all__ = ['EpubTests']

#: example html from static test-files directory
EXAMPLE_HTML = os.path.join(STATIC, 'example.html')

#: example book file-tree from static test-files directory
EXAMPLE_BOOK = os.path.join(STATIC, 'example_book')

#** Tests **#

class EpubTests(unittest.TestCase):
    """Epub Complete Book Generation UnitTests"""

    def setUp(self):
        """generate epub object for testing"""
        chapter   = create_chapter_from_file(EXAMPLE_HTML)
        self.epub = Epub('title')
        self.epub.add_chapter(chapter)

    def _walk(self, path: str) -> Set[str]:
        """generate a list of files included in tree"""
        tree = set()
        for root, _, files in os.walk(path):
            fpath = root.split(path, 1)[1]
            for file in files:
                tree.add(os.path.join(fpath, file))
        return tree

    def test_create_epub(self):
        """generate an epub directory-tree ready for zipping"""
        try:
            edir  = self.epub.build_epub_dir()
            tree1 = self._walk(edir)
            tree2 = self._walk(EXAMPLE_BOOK)
            self.assertEqual(tree1, tree2)
        finally:
            self.epub.builder.cleanup()
 
    def test_custom_cover(self):
        """ensure custom cover generation works as intended"""
        self.epub.cover = LOCAL_IMAGE
        try:
            edir  = self.epub.build_epub_dir()
            fname = os.path.basename(self.epub.cover)
            fpath = os.path.join(edir, f'OEBPS/images/{fname}')
            self.assertTrue(os.path.exists(fpath), 'no such cover')
        finally:
            self.epub.builder.cleanup()

#** Main **#

if __name__ == '__main__':
    unittest.main()



"""
Epub Object Definition and Generation
"""
from dataclasses import dataclass, field
from typing import Optional, List, Type

from .chapter import Chapter
from .builder import AssignedChapter, Assignment, EpubSpec, EpubBuilder

#** Variables **#
__all__ = ['Epub']

#** Classes **#

@dataclass
class Epub(EpubSpec):
    builder_factory: Type[EpubBuilder] = field(repr=False, default=EpubBuilder)

    def __post_init__(self):
        self.chapters: List[AssignedChapter] = []
        self.last_chapter = 1
        self.builder      = self.builder_factory(self)

    def assign_chapter(self) -> Assignment:
        """
        retrieve a valid assignment for an ebook chapter

        :return: chapter assignment information
        """
        id         = f'chapter_{self.last_chapter}'
        order      = self.last_chapter
        link       = f'chapter-{self.last_chapter}.xhtml'
        assignment = Assignment(id, link, order)
        self.last_chapter += 1
        return assignment

    def add_chapter(self, chapter: Chapter):
        """
        add a chapter to the ebook for later generation

        :param chapter: chapter to add to epub
        """
        assignment = self.assign_chapter()
        self.chapters.append((assignment, chapter))
        self.logger.debug(f'epub=%r added chapter %d: %r' % (
            self.title, assignment.play_order, chapter.title))

    def build_epub_dir(self) -> str:
        """
        generate and compile an uncompressed epub build directory

        :return: directory of epub content
        """
        dirs = self.builder.begin()
        for assign, chapter in self.chapters:
            self.builder.render_chapter(assign, chapter)
        self.builder.index()
        return dirs.basedir

    def create(self, fpath: Optional[str] = None) -> str:
        """
        generate an epub book at the given filepath

        :param fpath: filepath of finalized epub
        :return:      filepath of finished epub
        """
        try:
            self.build_epub_dir()
            return self.builder.compress(fpath)
        finally:
            self.builder.cleanup()



"""
Pypub Unit Tests
"""
import os

#** Variables **#
__all__ = ['ChapterTests', 'FactoryTests', 'EpubTests']

#: static testing files directory
STATIC = os.path.realpath(os.path.join(os.path.dirname(__file__), 'static'))

#: sample local file image
LOCAL_IMAGE = os.path.join(STATIC, 'sample.png')

#** Imports **#
from .epub import EpubTests
from .chapter import ChapterTests
from .factory import FactoryTests



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    readme = f.read()

setup(
    name='pypub3',
    version='2.0.8',
    license='MIT',
    author='Andrew Scott',
    author_email='imgurbot12@gmail.com',
    url='https://github.com/imgurbot12/pypub',
    description="A python3 library to generate custom epub books.",
    long_description=readme,
    long_description_content_type="text/markdown",
    python_requires='>=3.6',
    packages=find_packages(),
    install_requires=[
        'pyxml3>=0.0.2',
        'jinja2>=3.1.2',
        'dataclasses>=0.6',
        'pillow>=10.0.0',
        'filetype>=1.2.0'
    ],
    package_data={
        'pypub': [
            'templates/*',
            'static/*',
            'static/css/*',
            'static/img/*',
            'static/fonts/*'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



"""
Chapter Assignment and Processing Factories
"""
import html
import os.path
import urllib.request
from io import BytesIO
from dataclasses import dataclass
from typing import Optional

import pyxml.html

try:
    import mammoth
except ModuleNotFoundError:
    mammoth = None

#** Variables **#
__all__ = [
    'Chapter',

    'create_chapter_from_html',
    'create_chapter_from_text',
    'create_chapter_from_file',
    'create_chapter_from_url',
]

#: user agent to use when making url requests
user_agent = 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.81 Safari/537.36'

#** Classes **#

@dataclass(repr=False)
class Chapter:
    title:   str
    content: bytes
    url:     Optional[str] = None

    def __repr__(self) -> str:
        return 'Chapter(title=%r, url=%r, content=%d)' % (
            self.title, self.url, len(self.content))

#** Functions **#

def urlrequest(url: str, timeout: int = 10):
    """
    complete a url-request to the specified url
    """
    headers = {'User-Agent': user_agent}
    req = urllib.request.Request(url, headers=headers)
    return urllib.request.urlopen(req, timeout=timeout)

def htmltostring(root: pyxml.html.HtmlElement) -> bytes:
    """
    convert html to bytes
    """
    html = pyxml.html.tostring(root, method='xml')
    return html.encode() if isinstance(html, str) else html

def convert_text(text: str) -> bytes:
    """
    convert basic text into paragraphed html

    :param text: raw bytes text
    :return:     html version of text
    """
    root = pyxml.html.Element('body')
    for line in text.splitlines():
        line = line.strip()
        if not line:
            continue
        elem = pyxml.html.Element('p')
        elem.text = html.escape(line)
        root.append(elem)
    return htmltostring(root)

def convert_docx(data: bytes) -> bytes:
    """
    convert docx content into valid html

    :param data: docx archive bytes
    :return:     valid page html
    """
    if mammoth is None:
        raise ImportError(
            'cannot process `docx` without `mammoth`. Please pip install it')
    buffer = BytesIO(data)
    result = mammoth.convert_to_html(buffer)
    return result.value.encode()

def convert_content(path: str, data: bytes) -> bytes:
    """
    convert content according to the path extension

    :param path: path associated w/ data to denote data type
    :param data: raw bytes to process into html
    :return:     data processed into valid html
    """
    if path.endswith('.txt') or path.endswith('.text'):
        text = data.decode()
        return convert_text(text)
    elif path.endswith('.docx'):
        return convert_docx(data)
    return data

def create_chapter_from_html(
    html:          bytes,
    title:         Optional[str] = None,
    url:           Optional[str] = None,
    title_xpath:   Optional[str] = None,
    content_xpath: Optional[str] = None,
) -> Chapter:
    """
    generate a chapter object from the given html string

    :param html:          html string making up chapter content
    :param title:         title used for the given chpater
    :param url:           url assigned to this particular chapter
    :param title_xpath:   xpath used to find title in html
    :param content_xpath: xpath used to find content in html
    :return:              generated chapter object
    """
    html  = b'<html>' + html + b'</html>'
    etree = None
    # assign title according to title-xpath if specified
    if not title:
        etree = etree or pyxml.html.fromstring(html)
        xpath = (title_xpath or './/title').rsplit('/text()', 1)[0] + '/text()'
        elem  = etree.xpath(xpath)
        if not elem and title_xpath:
            raise ValueError(f'no such title: {title_xpath!r}')
        title = elem[0] if elem else 'Epub Chapter'
    # assign content according to content-xpath if specified
    if content_xpath:
        etree = etree or pyxml.html.fromstring(html)
        elem  = etree.xpath(content_xpath)
        root  = None
        if not len(elem):
            raise ValueError(f'no content at xpath: {content_xpath!r}')
        if len(elem) > 1:
            root = pyxml.html.Element('div')
            for child in elem:
                root.append(child)
        else:
            root = elem[0]
        html = htmltostring(root)
    # generate chapter object
    return Chapter(title, html, url)

def create_chapter_from_text(
    text:    str,
    title:   Optional[str] = None,
    url:     Optional[str] = None,
) -> Chapter:
    """
    generate a chapter object from the given text

    :param text:    text content to parse into chapter
    :param title:   assigned title of the chapter
    :param url:     url associated w/ the chapter
    """
    html = convert_text(text)
    return create_chapter_from_html(html, title, url)

def create_chapter_from_file(
    fpath:         str,
    title:         Optional[str] = None,
    url:           Optional[str] = None,
    title_xpath:   Optional[str] = None,
    content_xpath: Optional[str] = None,
) -> Chapter:
    """
    generate a chapter object from the given file

    :param fpath:         html file location making up chapter content
    :param title:         title used for the given chpater
    :param url:           url assigned to this particular chapter
    :param title_xpath:   xpath used to find title in html
    :param content_xpath: xpath used to find content in html
    """
    with open(fpath, 'rb') as f:
        url  = url or f'file://{os.path.abspath(fpath)}'
        html = convert_content(fpath, f.read())
        return create_chapter_from_html(html,
            title, url, title_xpath, content_xpath)

def create_chapter_from_url(
    url:           str,
    title:         Optional[str] = None,
    title_xpath:   Optional[str] = None,
    content_xpath: Optional[str] = None,
) -> Chapter:
    """
    generate a chapter object from the given file

    :param url:           url being downloaded for html content
    :param title:         title used for the given chpater
    :param title_xpath:   xpath used to find title in html
    :param content_xpath: xpath used to find content in html
    :param factory:       chapter factory override (for customization)
    """
    res  = urlrequest(url, timeout=10)
    html = convert_content(url, res.read())
    return create_chapter_from_html(html,
        title, url, title_xpath, content_xpath)



"""
Chapter Factory Unit Tests
"""
import os
import logging
import tempfile
from typing import Optional
import unittest

from . import LOCAL_IMAGE
from ..chapter import create_chapter_from_html, htmltostring
from ..factory import RenderCtx, SimpleChapterFactory

#** Variables **#
__all__ = ['FactoryTests']


#: basic testing logging instance
logger = logging.getLogger('pypub.test')

#: chapter factory instance used to generate chapters
factory = SimpleChapterFactory()

image_html_sample1 = b"""
<!DOCTYPE html>
<html>
    <head>
    </head>
    <body>
        <img src=""></img>
    </body>
</html>
"""

image_html_sample2 = b"""
<!DOCTYPE html>
<html>
    <head>
    </head>
    <body>
        <img src=""></img>
        <img></img>
        <img/>
    </body>
</html>
"""

article_html_sample1 = b"""
<html>
    <head></head>
    <body>
        <article>Hello! I am a test</article>
    </body>
</html>
"""

article_html_sample2 = b"""
<html>
    <head></head>
    <body>
        <div>dsfasfadfasdfasdf</div>
        <article>Hello! I am a test</article>
    </body>
</html>
"""

article_html_sample3 = b"""
<html>
    <head></head>
    <body>
        <article><video></video>Hello! I am a test</article>
    </body>
</html>
"""

complete_html_sample1 = b"""
<!DOCTYPE html>
<html>
    <head>
    </head>
    <body>
        <div>Hello </div>
    </body>
</html>
"""

complete_html_sample2 = b"""
<!DOCTYPE html>
<html>
    <head>
    </head>
    <body>
        <div>Hello </div>
        <script>Uh oh...it's an evil script!</script>
    </body>
</html>
"""

complete_html_sample3 = b"""
<!DOCTYPE html>
<html>
    <head>
    </head>
    <body>
        <div>Hello </div>
    </body>
    <video>Play me!</video>
</html>
"""

complete_html_sample4 = b"""
<!DOCTYPE html>
<html>
    <head>
    </head>
    <body>
        <video>
        <div>Hello </div>
        </video>
    </body>
    <video>Play me!</video>
</html>
"""

complete_html_sample5 = b"""
<!DOCTYPE html>
<html>
    <head>
    </head>
    <body>
        <video>
            <div>Hello&nbsp;</div>
        </video>
    </body>
    <video>Play me!</video>
</html>
"""

local_image_html_sample1 = f"""
<!DOCTYPE html>
<html>
    <head>
    </head>
    <body>
       <p>This is a test</p>
       <img src="file://{LOCAL_IMAGE}">
    </body>
</html>
""".encode()

#** Functions **#

def cleanup(content: bytes) -> bytes:
    """
    clean and produce finalized pypub chapter content
    """
    etree = factory.cleanup_html(content)
    for elem in etree.iter():
        elem.tail = None
        elem.text = elem.text.strip() if elem.text else ''
    return htmltostring(etree)

#** Tests **#

class FactoryTests(unittest.TestCase):
    """Epub Ebook Factory UnitTests"""

    def setUp(self):
        """create tempdir default of none"""
        self.tempdir: Optional[tempfile.TemporaryDirectory] = None

    def render_ctx(self, content: bytes) -> RenderCtx:
        """generate render context for more factory tests"""
        chapter = create_chapter_from_html(content)
        etree   = factory.cleanup_html(chapter.content)
        self.tempdir = tempfile.TemporaryDirectory()
        return RenderCtx(
            logger=logger, 
            chapter=chapter, 
            etree=etree, 
            imagedir=self.tempdir.name, 
            template=None, #type: ignore
        )

    def tearDown(self) -> None:
        """ensure teardown of temporary directory"""
        if self.tempdir is not None:
            self.tempdir.cleanup()

    def test_clean_images(self):
        """ensure etree cleaning works as intended to filter bad images"""
        clean1 = cleanup(image_html_sample1)
        clean2 = cleanup(image_html_sample2)
        self.assertEqual(clean1, clean2)

    def test_clean_article(self):
        """test that non-whitelisted tags get removed from html snippet"""
        s1 = cleanup(article_html_sample1)
        s2 = cleanup(article_html_sample2)
        s3 = cleanup(article_html_sample3)
        self.assertEqual(s1, s2)
        self.assertEqual(s1, s3)

    def test_clean_html(self):
        """test entire html grouping can be cleaned properly"""
        s1 = cleanup(complete_html_sample1)
        s2 = cleanup(complete_html_sample2)
        s3 = cleanup(complete_html_sample3)
        s4 = cleanup(complete_html_sample4)
        s5 = cleanup(complete_html_sample5)
        self.assertEqual(s1, s2)
        self.assertEqual(s1, s3)
        self.assertEqual(s1, s4)
        self.assertEqual(s1, s5)

    def test_local_image(self):
        """test local image resolution works as intended"""
        # attempt to hydrate local image
        ctx = self.render_ctx(local_image_html_sample1)
        factory.hydrate(ctx)
        # ensure image exists and is registered
        assert self.tempdir is not None
        src  = ctx.etree.find('//img/@src') or 'images/notfound.png'
        name = os.path.basename(src)
        fpath = os.path.join(self.tempdir.name, name)
        self.assertTrue(os.path.exists(fpath), 'local image not converted')

#** Main **#

if __name__ == '__main__':
    unittest.main()



"""
Epub library designed to build and generate ebooks in ePub format
"""

#** Variables **#
__all__ = [
    'Epub',
    'Assignment',
    'EpubBuilder',

    'Chapter',
    'ChapterFactory',
    'SimpleChapterFactory',

    'create_chapter_from_html',
    'create_chapter_from_text',
    'create_chapter_from_file',
    'create_chapter_from_url',
]

#** Imports **#
from .epub import Epub
from .builder import Assignment, EpubBuilder
from .chapter import *
from .factory import ChapterFactory, SimpleChapterFactory
